/* 
 * test_findpoca: This program is part of the liboctrope test suite. In
 *                particular, it attempts to exercise all the cases in the
 *                find_poca routines. 
 */

/* Copyright 2004 The University of Georgia. */

/* This file is part of liboctrope.
   
liboctrope is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

liboctrope is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with liboctrope; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif

#include "octrope.h"

int NUM_PASSED = 0;

void find_poca(octrope_vector edgeA[2], octrope_vector edgeB[2],
               double *s, double *t, double *dist);

void test_case(octrope_vector edgeA[2], octrope_vector edgeB[2], 
	       double s_expected, double t_expected, double dist_expected)

     /* Tests this case, exits with code 1 and prints error message if fails,
	returns silently if passed. */

{
  double s,t,dist;

  find_poca(edgeA,edgeB,&s,&t,&dist);

  if (octrope_error_num != 0) { /* should never happen but will stop the
                                   compiler warnings */
    fprintf(stderr,"test_findpoca: %s",octrope_error_str);
    exit(-1);
  }

  if (fabs(s - s_expected) > 1e-10 || 
      fabs(t - t_expected) > 1e-10 || 
      fabs(dist - dist_expected) > 1e-10) {

    fprintf(stderr,
	    "FAILED test_findpoca case.\n"
	    "EdgeA              EdgeB            \n"
	    "(%5g,%5g,%5g)      (%5g,%5g,%5g)    \n"
	    "(%5g,%5g,%5g)      (%5g,%5g,%5g)    \n"
	    "\n"
	    "s returned: %5g\t t returned: %5g\t dist returned: %5g\n"
	    "s expected: %5g\t t expected: %5g\t dist expected: %5g\n",
	    edgeA[0].c[0],edgeA[0].c[1],edgeA[0].c[2],
	    edgeB[0].c[0],edgeB[0].c[1],edgeB[0].c[2],
	    edgeA[1].c[0],edgeA[1].c[1],edgeA[1].c[2],
	    edgeB[1].c[0],edgeB[1].c[1],edgeB[1].c[2],
	    s,t,dist,s_expected,t_expected,dist_expected);
    exit(1);

  }

  NUM_PASSED++;

}
  
void poca_ordinary_correct(int times_to_run) {

  octrope_vector dirA = {{1,0,0}}, dirB = {{1,1,0}};
  octrope_vector startA = {{0,0,0}}, startB = {{0,0,1}};
  octrope_vector edgeA[2], edgeB[2];
  
  int i;
  

  /* There's nothing particularly intelligent about this code. 
     We list a bunch of cases for which we know the solutions, 
     and check the results of find_poca. */

  for(i=0;i<times_to_run;i++) {
    
    /* s,t in range. */
    
    edgeA[0] = octrope_vplus(startA,octrope_scalarmult(-1,dirA));
    edgeA[1] = octrope_vplus(startA,octrope_scalarmult(1,dirA));
    
    edgeB[0] = octrope_vplus(startB,octrope_scalarmult(-1,dirB));
    edgeB[1] = octrope_vplus(startB,octrope_scalarmult(1,dirB));
    
    /* We expect the closest point to be at s=0.5, t=0.5, dist = 1.*/
    
    test_case(edgeA,edgeB,0.5,0.5,1);
    
    /* s in range, t = 0. */
    
    edgeB[0] = octrope_vplus(startB,octrope_scalarmult(0.5,dirB));
    edgeB[1] = octrope_vplus(startB,octrope_scalarmult(3,dirB));
    
    test_case(edgeA,edgeB,0.75,0,sqrt(1.25)); 
    
    /* s in range, t = 1. */
    
    edgeB[0] = octrope_vplus(startB,octrope_scalarmult(-3,dirB));
    edgeB[1] = octrope_vplus(startB,octrope_scalarmult(-0.5,dirB));
    
    test_case(edgeA,edgeB,0.25,1,sqrt(1.25)); 
    
    /* t in range, s = 0. */
    
    edgeB[0] = octrope_vplus(startB,octrope_scalarmult(-1,dirB));
    edgeB[1] = octrope_vplus(startB,octrope_scalarmult(1,dirB));
    
    edgeA[0] = octrope_vplus(startA,octrope_scalarmult(0.5,dirA));
    edgeA[1] = octrope_vplus(startA,octrope_scalarmult(3,dirA));
    
    test_case(edgeA,edgeB,0,0.625,sqrt(9.0/8.0)); 
    
    /* t in range, s = 1. */
    
    edgeA[0] = octrope_vplus(startA,octrope_scalarmult(-3,dirA));
    edgeA[1] = octrope_vplus(startA,octrope_scalarmult(-0.5,dirA));
    
    test_case(edgeA,edgeB,1,0.375,sqrt(9.0/8.0)); 
    
    /* We now test the corners. */
    
    /* s = 0, t = 0 */
    
    edgeB[0] = octrope_vplus(startB,octrope_scalarmult(0.5,dirB));
    edgeB[1] = octrope_vplus(startB,octrope_scalarmult(3,dirB));
    
    edgeA[0] = octrope_vplus(startA,octrope_scalarmult(0.5,dirA));
    edgeA[1] = octrope_vplus(startA,octrope_scalarmult(3,dirA));
    
    test_case(edgeA,edgeB,0,0,sqrt(1.25)); 
    
    /* s = 1, t = 1 */
    
    edgeB[0] = octrope_vplus(startB,octrope_scalarmult(-3,dirB));
    edgeB[1] = octrope_vplus(startB,octrope_scalarmult(-0.5,dirB));
    
    edgeA[0] = octrope_vplus(startA,octrope_scalarmult(-3,dirA));
    edgeA[1] = octrope_vplus(startA,octrope_scalarmult(-0.5,dirA));
    
    test_case(edgeA,edgeB,1,1,sqrt(1.25)); 
    
    /* s = 0, t = 1 */
    
    /* These cases require a little different geometry. */
    
    edgeB[0] = octrope_vplus(startB,octrope_scalarmult(-3,dirB));
    edgeB[1] = octrope_vplus(startB,octrope_scalarmult(-0.5,dirB));
    
    edgeA[0] = octrope_vplus(startA,octrope_scalarmult(0,dirA));
    edgeA[1] = octrope_vplus(startA,octrope_scalarmult(1,dirA));
    
    test_case(edgeA,edgeB,0,1,sqrt(1.5)); 
    
    /* s = 1, t = 0 */
    
    edgeB[0] = octrope_vplus(startB,octrope_scalarmult(0.5,dirB));
    edgeB[1] = octrope_vplus(startB,octrope_scalarmult(3,dirB));
    
    edgeA[0] = octrope_vplus(startA,octrope_scalarmult(-1,dirA));
    edgeA[1] = octrope_vplus(startA,octrope_scalarmult(0,dirA));
    
    test_case(edgeA,edgeB,1,0,sqrt(1.5)); 
    
    /* (-8,0,2)--(-1,0,2), (0,-8,0)--(0,-1,0), expect s=1, t=1 */
    edgeA[0].c[0] = -8; edgeA[0].c[1] = 0; edgeA[0].c[2] = 2;
    edgeA[1].c[0] = -1; edgeA[1].c[1] = 0; edgeA[1].c[2] = 2;
    edgeB[0].c[0] = 0; edgeB[0].c[1] = -8; edgeB[0].c[2] = 0;
    edgeB[1].c[0] = 0; edgeB[1].c[1] = -1; edgeB[1].c[2] = 0;
    
    test_case(edgeA,edgeB,1,1,sqrt(6));

  }
}

void poca_parallel_correct() {

  octrope_vector dirA = {{1,0,0}};
  octrope_vector startA = {{0,0,0}}, startB = {{0,0,1}};
  octrope_vector edgeA[2], edgeB[2];

   /* The two segments overlap. */

  edgeB[0] = octrope_vplus(startB,octrope_scalarmult(0,dirA));
  edgeB[1] = octrope_vplus(startB,octrope_scalarmult(2,dirA));

  edgeA[0] = octrope_vplus(startA,octrope_scalarmult(-1,dirA));
  edgeA[1] = octrope_vplus(startA,octrope_scalarmult(1,dirA));
    
  test_case(edgeA,edgeB,0.75,0.25,1.0); 

  /* Edge B is to the right of edge A. */

  edgeB[0] = octrope_vplus(startB,octrope_scalarmult(2,dirA));
  edgeB[1] = octrope_vplus(startB,octrope_scalarmult(3,dirA));

  edgeA[0] = octrope_vplus(startA,octrope_scalarmult(0,dirA));
  edgeA[1] = octrope_vplus(startA,octrope_scalarmult(1,dirA));
    
  test_case(edgeA,edgeB,1,0,sqrt(2)); 

  /* Edge B is to the left of edge A. */

  edgeB[0] = octrope_vplus(startB,octrope_scalarmult(0,dirA));
  edgeB[1] = octrope_vplus(startB,octrope_scalarmult(1,dirA));

  edgeA[0] = octrope_vplus(startA,octrope_scalarmult(2,dirA));
  edgeA[1] = octrope_vplus(startA,octrope_scalarmult(3,dirA));
    
  test_case(edgeA,edgeB,0,1,sqrt(2)); 

}

double performance_test() 
{

  int NUM_TESTS = 100000;
  clock_t timeused,starttime;
  double  performance;

  fprintf(stderr,"Now making a performance test of find_poca.\n");

  starttime = clock();
  poca_ordinary_correct(NUM_TESTS);
  timeused = clock();

  performance = 13.0*(double)(NUM_TESTS)*(double)(CLOCKS_PER_SEC)/
    (double)(timeused - starttime);

  fprintf(stderr,"Approximately %g pocas per second on this machine.\n"
	  ,performance);
  exit(0);
 
}


int main() {

  poca_ordinary_correct(1);
  poca_parallel_correct();

  exit(0);

}


